--- Box2D plugin for STI
-- @module box2d
-- @author Landon Manning
-- @author Claude Marinier
-- @copyright 2021
-- @license MIT/X11

local love  = _G.love
local utils = require((...):gsub('plugins.box2d', 'utils'))
local lg    = require((...):gsub('plugins.box2d', 'graphics'))

return {
	box2d_LICENSE     = "MIT/X11",
	box2d_URL         = "https://github.com/karai17/Simple-Tiled-Implementation",
	box2d_VERSION     = "2.3.2.7",
	box2d_DESCRIPTION = "Box2D hooks for STI.",

	--- Initialize Box2D physics world.
    --
	-- @param world  The Box2D world to add objects to.
    --
	box2d_init = function(map, world)

		assert(love.physics, "To use the Box2D plugin, enable the love.physics module.")

        -- this is the main body, it defaults to static
        -- its position (centre) is set to map offset
        -- but the map offset seems to be the top left corner
        -- What is going on here?
        --
		local body      = love.physics.newBody(world, map.offsetx, map.offsety)

        -- keep a list of all collisions
        --
        -- box2d_draw debug function draws these collision "boxes"
        --
		local collision = {
			body = body,
		}

        -- keep a list of tile instances with the associated dynamic object
        --
        local dynamicTiles = { }

        --- test for empty table
        --
        -- @param aTable  A table to check
        --
        local function tableIsEmpty( aTable )
            -- recommended for efficiency, not sure it is needed
            local next = next  
            return next(aTable) == nil
            -- another suggestion
--          return type(next(aTable)) == "nil"
        end

        --- dump table recursively, but only so deep
        --
        -- @param obj    the object which could be a table
        -- @param max    maximum depth or 2 if not specified
        -- @param depth  used to control recursion
        --
        local function dumpTableFew(obj,max,depth)
            max = max or 2
            depth = depth or 1
            if type(obj) == 'table' and depth < max then
                local s = '{ '
                for k,v in pairs(obj) do
                    if type(k) ~= 'number' then k = '"'..k..'"' end
                    s = s .. '['..k..'] = ' .. dumpTableFew(v,max,depth+1) .. ', '
                end
                return s .. '} '
            else
                return tostring(obj)
            end
        end

        --- getPolygonVertices
        --
        -- @param object  must have a polygon entry
        --
		local function getPolygonVertices(object)
			local vertices = {}
			for _, vertex in ipairs(object.polygon) do
				table.insert(vertices, vertex.x)
				table.insert(vertices, vertex.y)
			end
			return vertices
		end

        --- setFixtureProperties
        --
        -- Set some custom fixture properties from userdata
        -- (or use default set by box2d)
        --
        -- @param fixture   the fixture, adding to this
        -- @param userdata  the user data
        --
        local function setFixtureProperties(fixture, userdata)
			fixture:setUserData(userdata)
			fixture:setFriction(userdata.properties.friction       or 0.2)
			fixture:setRestitution(userdata.properties.restitution or 0.0)
			fixture:setSensor(userdata.properties.sensor           or false)
			fixture:setFilterData(
				userdata.properties.categories or 1,
				userdata.properties.mask       or 65535,
				userdata.properties.group      or 0
			)
        end

        --- addObjectToWorld
        --
        -- create a new body for a collidable object which is not a tile
        --
        -- collidable tiles should be static and attached to the main body
        -- (the map), but I marked them as static on the map
        --
        -- After an object is converted to triangles, all the triangles
        -- must be kept together. How can I do this?
        --
        -- @param obj           the object I am adding
        -- @param userdata      user data
        -- @param tileOrObject  not used, stored in the collision table
        --
        -- Always return currentBody even though it will be used only
        -- for dynamic objects.
        --
		local function addObjectToWorld(obj, userdata, tileOrObject)

--[[
            print( "\naddObjectToWorld" )
            print( "  obj x     = " .. obj.x )
            print( "  obj y     = " .. obj.y )
            print( "  obj shape = " .. obj.shape )
            print( "  #vertices = " .. #obj.vertices )
--]]

            -- start with the body, there is only one
            --
            -- may attach many shapes to it
            --
			local currentBody

			-- dynamic is used for objects or players
            --
			if userdata.properties.dynamic == true then
                -- print("use new dynamic body")
				currentBody = love.physics.newBody(world, obj.x, obj.y, 'dynamic')

			-- static is used for things that shouldn't move, like walls/ground
            --
			elseif userdata.properties.static == true then
                -- print("use new static body")
				currentBody = love.physics.newBody(world, obj.x, obj.y, 'static')

			-- kinematic objects are static but effect other bodies
            --
			elseif userdata.properties.kinematic == true then
                -- print("use new kinematic body")
				currentBody = love.physics.newBody(world, obj.x, obj.y, 'kinematic')

            -- everything else is attached to the main body
            --
            else
                -- when this is used, everything is piled
                -- in the top left corner
                --
                -- print("use main body")
			    currentBody = body
			end

            -- now shape and fixture
            --
            -- will trangulate polygons and attach them to the same body
            --
			local shape
			local fixture

            -- this code assumes that everything that is not a polyline
            -- has been converted to a polygon
            --
			if obj.shape == "polyline" then

                -- check the Box2D documentation: Collision, Shapes
                --   - edges are two-sided and have only two points
                --   - chains are one-sided: watch vertex order (this
                --     would affect the way I draw collision shapes
                --     in Tiled)
                --
                -- It is working well: collisions work as expected. Did
                -- I accidentally make the chains in the correct order
                -- or does the one-sided feature only apply when looping?
                --
                -- I could check for two points (4 elements in vertex array)
                -- and use a newEdgeShape
				--   shape = love.physics.newEdgeShape(unpack(obj.vertices))
                --
				shape = love.physics.newChainShape(false,unpack(obj.vertices))

                fixture = love.physics.newFixture(currentBody, shape)
                setFixtureProperties(fixture, userdata)

                -- add it to the collision table
                --
                table.insert(collision, {
                                            object  = tileOrObject,
                                            body    = currentBody,
                                            shape   = shape,
                                            fixture = fixture,
                                        })

            elseif obj.shape == "polygon" then

                -- Elipses were converted to polygons in calculateObjectPosition
                --
                -- make shape and fixture
                --
                shape = love.physics.newPolygonShape(unpack(obj.vertices))
                fixture = love.physics.newFixture(currentBody, shape)
                setFixtureProperties(fixture, userdata)

                -- add it to the collision table
                --
                table.insert(collision, {
                                            object  = tileOrObject,
                                            body    = currentBody,
                                            shape   = shape,
                                            fixture = fixture,
                                        })
--[[
                -- Why are these converted to triangles?
                --
				local triangles = love.math.triangulate(obj.vertices)

				for _, triangle in ipairs(triangles) do

--                  print( "  triangle:", dumpTableFew(triangle,2) )

                    -- make shape and fixture for each
                    --
                    shape = love.physics.newPolygonShape(unpack(triangle))
                    fixture = love.physics.newFixture(currentBody, shape)
                    setFixtureProperties(fixture, userdata)

                    -- add it to the collision table
                    --
                    table.insert(collision, {
                                                object  = tileOrObject,
                                                body    = currentBody,
                                                shape   = shape,
                                                fixture = fixture,
                                            })
				end
--]]
			else

				shape = love.physics.newPolygonShape(unpack(obj.vertices))
                fixture = love.physics.newFixture(currentBody, shape)
                setFixtureProperties(fixture, userdata)

                -- add it to the collision table
                --
                table.insert(collision, {
                                            object  = tileOrObject,
                                            body    = currentBody,
                                            shape   = shape,
                                            fixture = fixture,
                                        })
			end

            return currentBody
		end

        --- calculateObjectPosition
        --
        -- prepare to add objects to the world
        --
        -- there is a lot happening in this function
        -- figure it out and document more
        --
        -- The addObjectToWorld function always returns the current
        -- body; we need to pass it on. It is used in the main loop
        -- for dynamic objects.
        --
        -- @param object
        -- @param tile
        --
		local function calculateObjectPosition(object, tile)

			local o = {
				shape   = object.shape,
				x       = (object.dx or object.x) + map.offsetx,
				y       = (object.dy or object.y) + map.offsety,
				w       = object.width,
				h       = object.height,
				polygon = object.polygon or object.polyline or object.ellipse or object.rectangle
			}

--[[
            print( "\ncalculateObjectPosition" )
            print( "  object:", dumpTableFew(object,2) )
            print( "  o:     ", dumpTableFew(o,2) )
--]]

            -- prepare user data
            -- sometimes the object has no properties
            -- that is OK because often the tile does
            --
			local userdata = {
				object     = o,
			}

            if object.properties ~= nil and
               not tableIsEmpty(object.properties)
            then
                userdata.properties = object.properties
            else
                userdata.properties = tile.properties
            end

			o.r = object.rotation or 0

			if o.shape == "rectangle" then

				local cos = math.cos(math.rad(o.r))
				local sin = math.sin(math.rad(o.r))
				local oy  = 0

                -- this is a rectangle represented as a polygon
                -- (points relative to object)
                --
				o.polygon = {
					{ x=0,   y=0   },
					{ x=o.w, y=0   },
					{ x=o.w, y=o.h },
					{ x=0,   y=o.h }
				}

                -- apply rotation
                --   change the polygon vertices in-place
                --
				for _, vertex in ipairs(o.polygon) do
					vertex.x, vertex.y =
                        utils.rotate_vertex(map, vertex, o.x, o.y, cos, sin, oy)
				end

				o.vertices = getPolygonVertices(o)
				return addObjectToWorld(o, userdata, tile or object)

--[[
        this is not going to work, need to be relative to the object

			elseif o.shape == "ellipse" then

                -- this code is not rotating, polygon will do it
                --
				if not o.polygon then
					o.polygon = utils.convert_ellipse_to_polygon(o.x, o.y, o.w, o.h)
				end

                -- Change it to a polygon and call back in.
                --
                o.shape = "polygon"
                calculateObjectPosition(o, tile)
--]]

			elseif o.shape == "polygon" then

				-- Recalculate collision polygons inside tiles
                --
                --   Why?
                --
				if tile then
					local cos = math.cos(math.rad(o.r))
					local sin = math.sin(math.rad(o.r))
					for _, vertex in ipairs(o.polygon) do
						vertex.x, vertex.y =
                            utils.rotate_vertex(map, vertex, o.x, o.y, cos, sin)
					end
				end

				o.vertices = getPolygonVertices(o)
                return addObjectToWorld(o, userdata, tile or object)

			elseif o.shape == "polyline" then

				o.vertices = getPolygonVertices(o)
				return addObjectToWorld(o, userdata, tile or object)

			end

            -- note that the above if statement ends with either a
            -- recursive call or a return statement
            --
            -->> We should never reach this point!

		end


        -- finished defining functions
        -- ===========================

        -- process tiles
        --
        -- each entry in the tileInstances table
        -- is a tile gid and a table of places where it appears
        -- (those are the instances)
        --
        -- each tile instance contains
        --   id      5        ; an instance of a tile
        --   gid     13       ; the tile
        --   layer   table    ; the layer where the tile is used
        --   batch   SpriteBatch
        --   x       128      ; position x
        --   y       0        ; position y
        --   r       0        ; rotation
        --   oy      0        ; not sure what this does
        --
        -- STI draws everything on all the layers. Dynamic objects should
        -- be on the object layer and marked as dynamic. After the physics
        -- engine does an update, we must update the position and orientation
        -- of all dynamic objects.
        -- There is a function to do this; it steps through the object layer
        -- looking for dynamic objects. Each one must include a reference to
        -- the matching physics body.
        -- This loop must identify dynamic objects and store the body in the
        -- map table.
        --
        for gid, tileInstance in pairs(map.tileInstances) do

            local tile = map.tiles[gid]
            local body

--          print( "\nprocessing this tile" )
--          print( dumpTableFew(tile,2) )
--          print( "tile properties" )
--          print( dumpTableFew(tile.properties,2) )

            -- Every place a tile is used (instanciated)
            --
            for i, instance in ipairs(tileInstance) do

--              print( "\nprocessing an instance/place" )
--              print( dumpTableFew(instance,2) )

                -- there could be more than one object on the tile
                --
                ---- Is the tile's collision polyline stored in
                ---- the object group? Check the map's Lua file.
                --
                if tile.objectGroup then
                    for _, object in ipairs(tile.objectGroup.objects) do
--                      print( "\nprocessing a tile object" )
--                      print( "gid: ", gid )
--                      print( " i : ", i )
--                      print( "object:", dumpTableFew(object,2) )
--                      print( "object properties:", dumpTableFew(object.properties,2) )
                        if object.properties.collidable == true or
                           tile.properties.collidable == true
                        then
                            -- Why a deep copy?
                            --
                            -- Especially since we add to it farther down.
                            -- Maybe to avoid changing it?
                            --
                            object = utils.deepCopy(object)
                            object.dx = instance.x + object.x
                            object.dy = instance.y + object.y
                            object.width = tile.width
                            object.height = tile.height

                            -- We now calculate object positions and add
                            -- objects to the physics world. We need to
                            -- update the position and orientation of
                            -- dynamic objects; save the body.
                            --
                            body = calculateObjectPosition(object, tile)

                            -- Save the body of this object only if it
                            -- is dynamic.
                            --
--                          if object.properties.dynamic == true or
--                             tile.properties.dynamic == true
--                          then
                            if tile.properties.dynamic == true then
                                -- Farther up, we make a deep copy. Why?
                                --
                                object.body = body
                                -- I need a list tying dynamic tiles and
                                -- their physics body.
                                -- I will want to update the position of the
                                -- tile's image where the drawing function will
                                -- see it.
                                --
--                              print("adding a tile-body pair")
                                table.insert(dynamicTiles, {
                                            tile = tile,
                                            body = body,
                                        })
                            end

                        end
                    end
--              else
--                  print( "tile ", gid, "does not have an objectGroup" )
--                  -- for example, a tile without a collision polyline
                end
            end

        end

		map.box2d_collision = collision
        map.box2d_dynamicTiles = dynamicTiles
	end,


    -- Update dynamic objects. It must be called after the physics
    -- update has run and before calling the STI draw function.
    --
    box2d_update = function(map)

        -- One way to do this:
        --   - go through instances (many more static ones)
        --   - find those on the object layer
        --   - find dynamic objects (tiles)
        --   - copy position and orientation from the body
        --     to the object
        -- I could save the body in the instance

        -- Is there a more direct way to find all dynamic objects?
        -- I could use the collision table. No, does not point to tile.
        -- A collision table entry contains { object, body, shape, fixture, }
        --   - find dynamic objects
        --   - copy position and orientation from the body
        --     to the object

        -- ******* ******* ******* ******* ******* ******* *******
        --
        -- try a special purpose table
        --
        -- does not work: collision box moves, but not image
        -- review options above
        -- I can see the positions and angles change
        --
--      io.write("\027[H\027[2J") -- clear screen
--      print("box2d_update")
        local x, y, r, ox, oy
		for _, dTile in ipairs(map.box2d_dynamicTiles) do
            x, y = dTile.body:getPosition()
            r = dTile.body:getAngle()
            ox = dTile.tile.objectGroup.objects[1].x
            oy = dTile.tile.objectGroup.objects[1].y
            lg.draw( map.tilesets[dTile.tile.tileset].image,
                     dTile.tile.quad,    x, y, r,
                     dTile.tile.sx, dTile.tile.sy, ox, oy)
--          print( x, y, r )
        end

    end,


	--- Remove Box2D fixtures and shapes from world.
	-- @param index The index or name of the layer being removed
    --
	box2d_removeLayer = function(map, index)
		local layer = assert(map.layers[index], "Layer not found: " .. index)
		local collision = map.box2d_collision

		-- Remove collision objects
		for i = #collision, 1, -1 do
			local obj = collision[i]

			if obj.object.layer == layer then
				obj.fixture:destroy()
				table.remove(collision, i)
			end
		end
	end,

	--- Draw Box2D collisions for debugging
    --
	-- @param tx Translate on X
	-- @param ty Translate on Y
	-- @param sx Scale on X
	-- @param sy Scale on Y
    --
	box2d_draw = function(map, tx, ty, sx, sy)

		local collision = map.box2d_collision

		lg.push()
		lg.scale(sx or 1, sy or sx or 1)
		lg.translate(math.floor(tx or 0), math.floor(ty or 0))

		for _, obj in ipairs(collision) do
			local points = {obj.body:getWorldPoints(obj.shape:getPoints())}
			local shape_type = obj.shape:getType()

			if shape_type == "edge" or shape_type == "chain" then
				love.graphics.line(points)
			elseif shape_type == "polygon" then
				love.graphics.polygon("line", points)
			else
				error("sti box2d_draw does not support "..shape_type.." shapes")
			end
		end

		lg.pop()
	end
}

--- Custom Properties in Tiled are used to tell this plugin what to do.
-- @table Properties
-- @field collidable set to true, can be used on any Layer, Tile, or Object
-- @field sensor set to true, can be used on any Tile or Object that is also collidable
-- @field dynamic set to true, can be used on any Tile or Object
-- @field static set to true, can be used on any Tile or Object
-- @field kinematic set to true, can be used on any Tile or Object
-- @field friction can be used to define the friction of any Object
-- @field restitution can be used to define the restitution of any Object
-- @field categories can be used to set the filter Category of any Object
-- @field mask can be used to set the filter Mask of any Object
-- @field group can be used to set the filter Group of any Object
